#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include "basic.h"

/* Memory allocation */
void *malloc_check(size_t sz) {
  if (sz == 0) return NULL ;
  void *p = malloc(sz) ;
  if (!p) {
    perror("malloc") ;
    exit(2) ;
  }
  return p ;
}

void *calloc_check(size_t nmemb,size_t sz) {
  if (sz == 0) return NULL ;
  void *p = calloc(nmemb,sz) ;
  if (!p) {
    perror("calloc") ;
    exit(2) ;
  }
  return p ;
}


/* Specific error function */

void errexit(char *msg,int err) {
  fprintf(stderr,"%s: %s\n",msg,strerror(err)) ;
  exit(2) ;
}

/* Boxed int */

void free_boxed_int(boxed_int_t *p) { free(p) ; }

boxed_int_t *alloc_boxed_int(int v) {
  boxed_int_t *r = malloc_check(sizeof(*r)) ;
  r->v = v ;
  return r ;
}


/* Threads */

void create_thread(pthread_t *th,void *(*f)(void *),void *x) {
  int errno ;
  if ((errno = pthread_create(th,NULL,f,x))) {
    errexit("pthread_create",errno) ;
  }
}

void create_thread_detached(pthread_t *th,void *(*f)(void *),void *x) {
  pthread_attr_t tattr;

  if (pthread_attr_init(&tattr)) {
    perror("pthread_attr_init") ;
    exit(2) ;
  }
  if (pthread_attr_setdetachstate(&tattr,PTHREAD_CREATE_DETACHED)) {
    perror("pthread_attr_setdetachstate") ;
    exit(2) ;
  }

  if (pthread_create(th,&tattr,f,x)) {
    perror("pthread_create") ;
    exit(2) ;
  }
  if (pthread_attr_destroy(&tattr)) {
    perror("pthread_attr_destroy") ;
    exit(2) ;
  }
}

void *join_thread(pthread_t *th) {
  void *r ;
  if (pthread_join (*th,&r)) {
    perror("pthread_join") ;
    exit(2) ;
  }
  return r ;
}
/* Mutex */

pthread_mutex_t *alloc_mutex(void) {
  pthread_mutex_t *r = malloc_check(sizeof(*r)) ;
  if (pthread_mutex_init(r,NULL)) {
    perror("pthread_mutex_init") ;
    exit(2) ;
  }
  return r ;
}

void free_mutex(pthread_mutex_t *p) {
  if (pthread_mutex_destroy(p)) {
    perror("pthread_mutex_destroy") ;
    exit(2) ;
  }
    free(p) ;
}

void lock_mutex(pthread_mutex_t *p) {
  if (pthread_mutex_lock(p)) {
    perror("pthread_mutex_lock") ;
    exit(2) ;
  }
}

void unlock_mutex(pthread_mutex_t *p) {
  if (pthread_mutex_unlock(p)) {
    perror("pthread_mutex_unlock") ;
    exit(2) ;
  }
}

/* Condition variable */

pthread_cond_t *alloc_cond(void) {
  pthread_cond_t *r = malloc_check(sizeof(*r)) ;
  if (pthread_cond_init(r,NULL)) {
    perror("pthread_cond_init") ;
    exit(2) ;
   }
  return r ;
}

void free_cond(pthread_cond_t *p) {
  if (pthread_cond_destroy(p)) {
    perror("pthread_cond_destroy") ;
    exit(2) ;
  }
  free(p) ;
}

void wait_cond(pthread_cond_t *c, pthread_mutex_t *m) {
  if (pthread_cond_wait(c,m)) {
    perror("pthread_cond_wait") ;
    exit(2) ;
  }
}

void signal_cond(pthread_cond_t *c) {
  if (pthread_cond_signal(c)) {
    perror("pthread_cond_signal") ;
    exit(2) ;
  }
}

void broadcast_cond(pthread_cond_t *c) {
  if (pthread_cond_broadcast(c)) {
    perror("pthread_cond_broadcast") ;
    exit(2) ;
  }
}

/* Barrier */

pthread_barrier_t *alloc_barrier(unsigned count) {
  pthread_barrier_t *p = malloc_check(sizeof(*p)) ;
  int err ;
  if ((err = pthread_barrier_init(p,NULL,count)))
    errexit("pthread_barrier_init",err) ;
  return p ;
}

void free_barrier(pthread_barrier_t *p) {
  int err ;
  if ((err = pthread_barrier_destroy(p)))
    errexit("pthread_barrier_destroy",err) ;
  free(p) ;
}


int wait_barrier(pthread_barrier_t *p) {
  int ret = pthread_barrier_wait(p) ;
  if (ret != 0 && ret != PTHREAD_BARRIER_SERIAL_THREAD)
    errexit("pthread_barrier_wait",ret) ;
  return ret ;
}


/* Delay */

void delay(unsigned int ms) {
  struct timespec req,rem ;
  unsigned int sec = ms / 1000 ;
  long ns = (ms % 1000) * 100000L ;
  req.tv_sec = sec ; req.tv_nsec = ns ;
  while (nanosleep(&req,&rem)) {
    if (errno == EINTR) {
      req = rem ;
    } else {
      errexit("nanosleep",errno) ;
    }
  }
}

/* Timing info */

tsc_t timeofday(void) {
  struct timeval tv ;
  if (gettimeofday(&tv,NULL)) {
    errexit("gettimeofday",errno) ;
  }
  return tv.tv_sec * ((tsc_t)1000000) + tv.tv_usec ;
}

double tsc_millions(tsc_t t) {
  return t / 1000000.0 ;
}
